@@ -1,13 +1,13 @@
-import unittest
-import tempfile
+from __future__ import unicode_literals
+
 import os
-import shutil
 import pickle
+import shutil
 import sys
-from io import StringIO
+import tempfile
+import unittest
 
 import pytest
-from mock import patch, mock_open
 
 import cloudpickle
 
@@ -19,7 +19,7 @@ class CloudPickleFileTests(unittest.TestCase):
     def setUp(self):
         self.tmpdir = tempfile.mkdtemp()
         self.tmpfilepath = os.path.join(self.tmpdir, 'testfile')
-        self.teststring = u'Hello world!'
+        self.teststring = 'Hello world!'
 
     def tearDown(self):
         shutil.rmtree(self.tmpdir)
@@ -99,18 +99,6 @@ class CloudPickleFileTests(unittest.TestCase):
         self.assertRaises(pickle.PicklingError,
                           lambda: cloudpickle.dumps(sys.stdin))
 
-    def NOT_WORKING_test_tty(self):
-        # FIXME: Mocking 'file' is not trivial... and fails for now
-        from sys import version_info
-        if version_info.major == 2:
-            import __builtin__ as builtins  # pylint:disable=import-error
-        else:
-            import builtins  # pylint:disable=import-error
-
-        with patch.object(builtins, 'open', mock_open(), create=True):
-            with open('foo', 'w+') as handle:
-                cloudpickle.dumps(handle)
-
 
 if __name__ == '__main__':
     unittest.main()
